/*
*   ProjectionGui
*
*   Step of the calibration which display point of view of the video camera and allows to delimit the border of the projection.
*   Since there is no drawing event, we have to add a widget (MaskEdit) which is generated with the frame and the borders drawn.
*   Each time we want to modify the borders, we will edit the variable containing the borders and trigger paint event of the widget
*   showing the camera with our projection and delimitations.
*/

#include "croppingmaskgui.h"
#include "ui_croppingmaskgui.h"

CroppingMaskGui::CroppingMaskGui(SandboxSetup *sandbox, MonitorGui *_mg, QWidget *parent) :
    SubApp("Cropping mask", "Error", parent),
    ui(new Ui::CroppingMaskGui)
{

    setup = sandbox;
    mg = _mg;
    ui->setupUi(this);
    ui->txtInstructions->setAlignment(Qt::AlignJustify);
    maskEdit = new MaskEdit(ui->frame->geometry());
    ui->vLayout->addWidget(maskEdit, 1);

    blueScreen = new QtFullScreen(mg->getResolution(), true, this);

    timer = new QTimer(this);
    connect(timer, &QTimer::timeout, this, &CroppingMaskGui::refreshFrame);
}

CroppingMaskGui::~CroppingMaskGui()
{
    delete blueScreen;
    delete maskEdit;
    delete ui;
}

void CroppingMaskGui::valideRoutine(){

    timer->stop();
    setup->getCamera()->capture();
    std::vector<cv::Point> rectPoints = getRectPoints();
    cv::Mat depthFrame = setup->getCamera()->getDepthFrame();
    
    setup->setupCroppingMask(rectPoints, rectPoints[0]);
    cv::Rect mask = setup->getCamera()->getCroppingMask();

    cv::Point2i centerProjection = cv::Point2i(mask.width/2, mask.height/2);
    setup->setupAdjustMatrix(rectPoints, cv::Point2i(0,0));
    setup->getProjection()->setDistanceTopSandbox(depthFrame.at<float>(centerProjection));
    endSuccess = true;
}

void CroppingMaskGui::cancelRoutine(){
    timer->stop();
    endSuccess = false;
}

void CroppingMaskGui::showEvent(QShowEvent *event){
    QWidget::showEvent(event);

    // show blue screen
    if(!blueScreenInitialized){
        blueScreenInitialized = true;
        blueScreen->editGeometry(mg->getResolution());
    }
    cv::Mat blueMat = cv::Mat(1, 1, CV_8UC3, cv::Scalar(0, 0, 255));
    blueScreen->imShow(blueMat);
    init();
    timer->start(100);
}

void CroppingMaskGui::closeEvent(QCloseEvent *event){
    QWidget::closeEvent(event);
    blueScreen->close();
}

/*
*   Initialize the frame displayed with the cropping mask by default or the one availible in the configuration
*/
void CroppingMaskGui::init(){

    setup->getCamera()->capture();
    cameraColoredFrame = setup->getCamera()->getColorFrame();

    // load from config
    if(rectPoints.empty()){
        if(!setup->loadCroppingMaskAndAdjustingMatrix()){
            cv::Rect mask = setup->getCamera()->getCroppingMask();
            cv::Point2i centerRotation = cv::Point2i(mask.x, mask.y);
            double angle = setup->getProjection()->getAngleRotation();
            std::vector<cv::Point2i> pts = { cv::Point2i(mask.x, mask.y),
                                             cv::Point2i(mask.x, mask.y+mask.height),
                                             cv::Point2i(mask.x+mask.width, mask.y+mask.height),
                                             cv::Point2i(mask.x+mask.width, mask.y) };
            for(cv::Point2i pt : pts){
                loadedMask.push_back(setup->getProjection()->revertRotatePixel(centerRotation, angle, pt));
            }
            rectPoints = loadedMask;
        }
    }

    // no config found
    if(!maskValideInFrame(&cameraColoredFrame)){
            float y = cameraColoredFrame.size().height;
            float x = cameraColoredFrame.size().width;
            rectPoints = std::vector<cv::Point2i>{ cv::Point2i(1.0/4*x, 1.0/4*y), cv::Point2i(1.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 1.0/4*y) };
    }

    maskEdit->updateFrame(&cameraColoredFrame, &rectPoints);
}

void CroppingMaskGui::refreshFrame(){
    setup->getCamera()->capture();
    cameraColoredFrame = setup->getCamera()->getColorFrame();
    maskEdit->updateFrame(&cameraColoredFrame);
}

/*
*   check if the mask from the configuration is valide in the frame from the camera
*/
bool CroppingMaskGui::maskValideInFrame(cv::Mat_<cv::Vec3b> *rgb){

    if(rectPoints.empty())
        return false;

    for(uint i=0; i<rectPoints.size(); i++){
        if( !(( 0<=rectPoints[i].x && rectPoints[i].x<rgb->size().width ) &&
            ( 0<=rectPoints[i].y && rectPoints[i].y<rgb->size().height ) ))
            {
            return false;
        }
    }

    return true;
}

/*
*   Set the mask with the default values
*/
void CroppingMaskGui::on_btnClear_clicked()
{
    if(cameraColoredFrame.size().height != 0 && cameraColoredFrame.size().width != 0){
        float y = cameraColoredFrame.size().height;
        float x = cameraColoredFrame.size().width;
        rectPoints = std::vector<cv::Point2i>{ cv::Point2i(1.0/4*x, 1.0/4*y), cv::Point2i(1.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 3.0/4*y), cv::Point2i(3.0/4*x, 1.0/4*y) };
        maskEdit->updateFrame(&rectPoints);
    }
}

/*
*   Set the mask with the one from the configuration (if any)
*/
void CroppingMaskGui::on_btnReset_clicked()
{
    if(!loadedMask.empty()){
        rectPoints = loadedMask;
        maskEdit->updateFrame(&rectPoints);
    }
}
