#ifndef SANDBOX_CAMERA_H
#define SANDBOX_CAMERA_H

#include <librealsense2/rs.hpp>
#include <librealsense2/rsutil.h>
#include <opencv2/opencv.hpp>

class Camera{
    private:
        // spatial filter, improve depth data accuracy
        rs2::spatial_filter *spatFilter;
        // temporal filter, improve depth data stability
        rs2::temporal_filter *tempFilter;
        // config can manage RealSense cameras and streams
        rs2::config *cfg;
        // manage the streams of cameras
        rs2::pipeline *pipe;
        // manage the stream's of color and depth frames based on the depth point of view
        rs2::align *align_to_depth;
        
        // frame objects from the camera
        rs2::video_frame *color_frame;
        rs2::depth_frame *depth_frame;
        
        // intrisic profil of the camera based on the objective and its configuration
        // (this won't be the same profil for the video camera or the depth camera
        // and depends on the configuration of the stream(resolution,...) too)
        rs2_intrinsics intr_profile;
        // factor to normalize the datas from depth frame to distances in meter
        float depth_scale;
        // Cropping mask (config parameter determined by the calibration), allows to set the ROI(range of interest) in an OpenCV Matrix
        cv::Rect croppingMask;

        void warmUpDepthLens();
        void filterDepthFrame(rs2::depth_frame &frameDepth);

    public:
        Camera();
        ~Camera();

        // Return values from depth matrix to real world (matrix of floats in meter)
        // the depths are the Z values from a 3D orthonormed system, where (0,0,0) is defined by the position of the camera
        // so it's not the distance from the camera to the point, it's its Z value
        cv::Mat getDepthFrame();
        cv::Mat getColorFrame();
        void setCroppingMask(cv::Rect mask){ croppingMask = mask; };
        cv::Rect getCroppingMask(){ return croppingMask; };
        
        int start();
        void capture();
        void stop();
        cv::Point3f deprojectPixelToPoint(float coord[], float z);
        cv::Point3f deprojectPixelToPoint(float coord[], float z, cv::Point2f fxy, cv::Point2f ppxy);
        cv::Point2i projectPointToPixel(cv::Point3f point3D);
        cv::Point2i projectPointToPixel(cv::Point3f point3D, cv::Point2f fxy, cv::Point2f ppxy);
        std::vector<cv::Point2f> getAdaptedIntrinsics(cv::Mat &projection);
        
        void printCroppingMask();

};
#endif