#ifndef SANDBOX_H
#define SANDBOX_H

#include <opencv2/opencv.hpp>
#include "camera.h"
#include "projection.h"
#include "beamer.h"
#include "sandboxConfig.h"

/*!
* @class Sandbox
* API to build augmented reality applications which use the configuration file
*/
class Sandbox{
    private:
        char *defaultConfigFilePath = (char *)"./sandbox_conf.yaml";
        Projection *projection;
        Camera *camera;
        Beamer *beamer;
        cv::Mat_<cv::Vec3b> projectedFrame;

    public:
        Sandbox();
        ~Sandbox();

        Camera* getCamera(){ return camera; };
        Beamer* getBeamer(){ return beamer; };
        Projection* getProjection(){ return projection; };

        /*!
        *   Initialize the components needed for the applications
        *   @return int indicating if there was an error, where 0 indicates no error 
        */
        int init();

        /*!
        *   Capture a frame from the camera, this update the depth buffer and the color buffer of the camera
        */
        void captureFrame();

        /*!
        *   Get a copy of the colored RGB frame from the buffer
        *   @return Colored frame represented by a matrix filled of 3 bytes values
        */
        cv::Mat_<cv::Vec3b> getColorFrame();

        /*!
        *   Get a copy of the depth frame from the buffer with values in meter
        *   @return Depth frame represented by a matrix filled of float values
        */
        cv::Mat_<float> getDepthFrame();

        /*!
        *   Adjust the frame to the topology by taking a picture with the depth camera
        *   @param frame Image to be adapted on the sandbox
        *   @return The image adjusted to the topology in RGB 
        */
        cv::Mat_<cv::Vec3b> adjustProjection(cv::Mat_<cv::Vec3b> &frame);

        /*!
        *   Adjust the frame to the topology by adapting to the topology passed with the frame
        *   @param frame Image to be adapted on the sandbox
        *   @param depth Depth frame representing the topology of the sand
        *   @return The image adjusted to the topology in RGB 
        */
        cv::Mat_<cv::Vec3b> adjustProjection(cv::Mat_<cv::Vec3b> &frame, cv::Mat_<float> &depth);

        /*!
        *   Load the configuration from the file at the default location
        *   @return int indicating if there was an error, where 0 indicates no error 
        */
        int loadConfig();

        /*!
        *   Load the configuration from the file at the specified location
        *   @param path to the configuration file
        *   @return int indicating if there was an error, where 0 indicates no error 
        */
        int loadConfig(char *path);
    
};

#endif
